#include "Heap.h"
#include "Heap_Impl.cpp"
#include <cassert>
#include <memory>
#include <iostream>

using namespace std;

namespace Combustion
{
	int g_NextAllocNum = 1;
	int GetAllocationNumber() { return g_NextAllocNum++; }

	Heap::Heap(void* pMemory, size_t size, int line /* = 0*/, char* filename /* = 0*/, int tag /* = 0*/)
	{
		if(!pMemory) throw MemoryException("Heap Initialized With Null Data", line, filename);
		if(size < sizeof(Heap)) throw MemoryException("Heap Size Too Small", line, filename);

		heap = new(pMemory) Heap_Impl(size);
		#ifdef _DEBUG
			heap->allocNum = GetAllocationNumber();
		#endif
	}

	Heap::~Heap()
	{
		heap->~Heap_Impl();
	}

#define FIND_SUITABLE_FREE while(current != 0) { \
			if(!current->isAllocated && current->size >= correctSize) break;\
			current = current->nextBlock; }

	void* Heap::Alloc(size_t size, int line /* = 0*/, char* filename /* = 0*/, int tag /* = 0*/)
	{
		size_t correctSize = size + GUARD_SIZE;//BLOCK_HEADER_AND_GUARD;

		//First-fit allocation (for speed)
		MemoryHeader* current = (MemoryHeader*) (reinterpret_cast<char*>(heap) + sizeof(Heap_Impl));

		FIND_SUITABLE_FREE

		if(!current)
		{
			Coalesce();	//if no suitable block found, coalesce to try to find one
			FIND_SUITABLE_FREE //try again
			if(!current) throw MemoryException("Out of Memory", line, filename);
		}

		heap->Split(current, size); //if no split is needed, none is made
		
		#ifdef _DEBUG
			heap->stats.numAllocs++;
			current->allocNum = GetAllocationNumber();
			current->lineNum = line;
			current->file = filename;
			current->tag = tag;
		#endif
		
		return reinterpret_cast<void*>( (reinterpret_cast<char*>(current) + sizeof(MemoryHeader)) );
	}

	Heap* Heap::SubHeap(size_t size, int line /* = 0*/, char* filename /* = 0*/)
	{
		//allocate enough room for everything	HeapObj Heap_Impl MemoryHeader size
		size_t correctSize = sizeof(Heap) + sizeof(Heap_Impl) + sizeof(MemoryHeader) + size;

		Heap* subheap = reinterpret_cast<Heap*>( Alloc(correctSize, line, filename) );
		void* subheapData = reinterpret_cast<void*>( reinterpret_cast<char*>(subheap) + sizeof(Heap) );
		subheap = new(subheap) Heap(subheapData, size, line, filename);
		
		return subheap;
	}

	void Heap::FreeSubHeap(Heap* subheap, int line /*= 0*/, char* filename /*= 0*/)
	{
		subheap->~Heap();					//delete subeap
		Free(subheap, line, filename);		//free data in outer heap
	}

	void Heap::Free(void* pMemory, int line /*= 0*/, char* filename /*= 0*/)
	{
		if(!pMemory) throw MemoryException("Null Pointer", line, filename);
		if(IsInvalid(pMemory)) throw MemoryException("Invalid Pointer in Free", line, filename);

		MemoryHeader* memHeader = (MemoryHeader*) (reinterpret_cast<char*>(pMemory) - sizeof(MemoryHeader));

		if(!memHeader->isAllocated)	throw MemoryException("Double Free", line, filename);

		memHeader->isAllocated = false;
		memHeader->guardData = _MAGIC;

		int* endGuard = (int*) (reinterpret_cast<char*>(pMemory) + memHeader->size);
		*endGuard = _MAGIC;

		memHeader->size += GUARD_SIZE; //the guard block now counts towards the block size

		#ifdef _DEBUG
			heap->stats.numDeallocs++;
		#endif

		heap->numAllocatedBlocks--;
		heap->currUsage -= memHeader->size;
	}

	void Heap::Coalesce(void *pMemory)
	{
		if(pMemory)
		{//only coalesce immediately surrounding blocks
			MemoryHeader* memHeader = (MemoryHeader*) (reinterpret_cast<char*>(pMemory) - sizeof(MemoryHeader));

			//walk left until a non-free block is hit
			MemoryHeader* start = memHeader;
			while(start->prevBlock != 0 && !start->prevBlock->isAllocated)
				start = start->prevBlock;

			//march right, coalescing until a non-free block is hit
			while(start->nextBlock != 0 && !start->nextBlock->isAllocated)
				heap->MergeWithNextBlock(start);
		}
		else
		{//coalesce all free blocks in the heap...
			MemoryHeader* memHeader = (MemoryHeader*) (reinterpret_cast<char*>(heap) + sizeof(Heap_Impl));

			while(memHeader != 0)
			{
				if(!memHeader->isAllocated && memHeader->nextBlock != 0 && !memHeader->nextBlock->isAllocated)
					heap->MergeWithNextBlock(memHeader);	//merge, but don't move memHeader
				else
					memHeader = memHeader->nextBlock; //don't merge, but move memHeader
			}
		}
	}

	size_t Heap::TotalMemory(bool checkFree)
	{
		//Walk through the memory and count the size of the free blocks
		MemoryHeader* current = (MemoryHeader*)(reinterpret_cast<char*>(heap) + sizeof(Heap_Impl));

		size_t totalFreeMem = 0;

		while(current != 0)
		{
			if( (checkFree && !current->isAllocated) || (!checkFree && current->isAllocated) )
				totalFreeMem += current->size;

			current = current->nextBlock;
		}

		return totalFreeMem;
	}

	size_t Heap::LargestFreeBlock()
	{
		//Walk through the memory and find the biggest free
		MemoryHeader* current = (MemoryHeader*)(reinterpret_cast<char*>(heap) + sizeof(Heap_Impl));

		size_t largestFree = 0;

		while(current != 0)
		{
			if(!current->isAllocated && current->size > largestFree)
				largestFree = current->size;

			current = current->nextBlock;
		}

		return largestFree;
	}

	void* Heap::OldestBlock()
	{
		#ifndef _DEBUG
				return 0;
		#else
			//Walk through the memory and find the earliest allocation
			MemoryHeader* current = (MemoryHeader*)(reinterpret_cast<char*>(heap) + sizeof(Heap_Impl));

			size_t oldest = INT_MAX;
			void* pOldest = 0;

			while(current != 0)
			{
				if(current->isAllocated && current->allocNum < oldest)
				{
					oldest = current->allocNum;
					pOldest = reinterpret_cast<void*>( reinterpret_cast<char*>(current) + sizeof(MemoryHeader));
				}

				current = current->nextBlock;
			}
			return pOldest;
		#endif
	}

	bool Heap::Validate()
	{
		//walk entire heap, checking each block
		MemoryHeader* memHeader = reinterpret_cast<MemoryHeader*>(
									(reinterpret_cast<char*>(heap) + sizeof(Heap_Impl)) );
		while(memHeader != 0)
		{
			if(memHeader->guardData != _MAGIC)
				return false;
			if(memHeader->isAllocated)
			{
				int* endGuard = (int*)(reinterpret_cast<char*>(memHeader) + sizeof(MemoryHeader) + memHeader->size);
					if(*endGuard != memHeader->guardData)
						return false;
			}

			memHeader = memHeader->nextBlock;
		}

		return true;
	}

	bool Heap::IsInvalid(void* pMemory)
	{
		if(pMemory == 0) return true;

		//check if the address is in the range of the heap
		void* start = reinterpret_cast<void*>(heap);
		void* end = (void*) (reinterpret_cast<char*>(heap) + sizeof(Heap_Impl) + heap->size);
		if(pMemory <= start || pMemory >= end)
			return true;

		MemoryHeader* memHeader = (MemoryHeader*) (reinterpret_cast<char*>(pMemory) - sizeof(MemoryHeader));

		//check to make sure the pointer was allocated
		if(!memHeader->isAllocated)
			return true;

		//check the guard blocks
		if(memHeader->guardData != _MAGIC)
			return true;
		if(memHeader->isAllocated)
		{
			int* endGuard = (int*)(reinterpret_cast<char*>(pMemory) + memHeader->size);
			if(*endGuard != memHeader->guardData)
				return true;
		}

		return false;
	}

	void Heap::PrintHeapStatistics()
	{
		cout << "===================================\n";
#ifdef _DEBUG
		cout << "Number of Allocations: " << heap->stats.numAllocs << endl;
		cout << "Number of Deallocations: " << heap->stats.numDeallocs << endl;
#endif
		cout << "Number of Blocks: " << heap->numTotalBlocks << endl;
		cout << "Number of Alloced: " << heap->numAllocatedBlocks << endl;
		cout << "Number of Free: " << heap->numTotalBlocks - heap->numAllocatedBlocks << endl;
		cout << "===================================\n" << endl;
	}

	float Heap::ExternalFragmentation()
	{
		int totalFreeBlocks = heap->numTotalBlocks - heap->numAllocatedBlocks;
		return 1.0f - (LargestFreeBlock() / totalFreeBlocks);
	}

	float Heap::InternalFragmentation()
	{
		return float(heap->numTotalBlocks * sizeof(MemoryHeader)) / float(heap->size);
	}

	size_t Heap::GetUsageFromTag(int tag)
	{
		#ifdef _DEBUG
				//Walk through the memory and count the size of the free blocks
				MemoryHeader* current = (MemoryHeader*)(reinterpret_cast<char*>(heap) + sizeof(Heap_Impl));
		
				size_t totalUsed = 0;
		
				while(current != 0)
				{
					if(current->isAllocated && current->tag == tag)
						totalUsed += current->size;
		
					current = current->nextBlock;
				}
		
				return totalUsed;
		#else
				return 0;
		#endif
	}

	int Heap::NumLeaks()
	{
		#ifdef _DEBUG
					return heap->stats.numAllocs - heap->stats.numDeallocs;//TODO!
		#else
				return false;
		#endif
	}

	void Heap::PrintLeaks()
	{
		#ifdef _DEBUG
			
		//walk entire heap, checking each block
		MemoryHeader* memHeader = reinterpret_cast<MemoryHeader*>(
									(reinterpret_cast<char*>(heap) + sizeof(Heap_Impl)) );
		while(memHeader != 0)
		{
			if(memHeader->isAllocated)
			{
				cout << "Leak in file " << memHeader->file << " on line " << memHeader->lineNum << ".\n";
				cout << "Allocation number: " << memHeader->allocNum << ".\n";
				if(memHeader->tag != 0) 
				{
					char tag[4];
					memcpy(tag, &(memHeader->tag), 4);
					cout << "Tagged As: ";
					for(int i = 3; i >= 0; --i)	//tag is backwards, so print in reverse
						cout << tag[i];
					cout <<"\n\n";
				}
			}

			memHeader = memHeader->nextBlock;
		}

		#endif
	}

	int Heap::GetMaxMemoryUsage()
	{
		return heap->maxUsage;
	}
}