#pragma once

#include "MemoryException.h"
#include "MemoryStructs.h"

namespace Combustion
{
	struct Heap_Impl;

	class Heap
	{
	public:
		/*	Creates a Heap that manages the provided memory
			pMemory: pointer to the memory to manage
			size:	 size of pMemory
			
			Optional Params (used for more verbose debugging data):
			line:	 line number of the call to create the heap
			filename:filename of the line with the call to create the heap
			tag:	 an integer representing a 4 character string to mark the memory

			throws MemoryException for failures
		*/
		Heap(void* pMemory, size_t size, int line = 0, char* filename = 0, int tag = 0);

		/*  Destroys the Heap
			Note: Does not call delete on the pMemory provided during the Heaps construction
		*/
		~Heap();

		/*	Allocates at least the amount of memory specified of the given type and returns a typed pointer
			size: Amount of memory requested
			
			Optional Params (used for more verbose debugging data):
			line:	 line number of the call to allocate memory
			filename:filename of the line with the call to allocate memory
			tag:	 an integer representing a 4 character string to allocate memory

			throws MemoryException for failures
		*/
		template<typename T>
		T* Alloc(size_t size, int line = 0, char* filename = 0, int tag = 0)
		{ return reinterpret_cast<T*>( Alloc(size, line, filename, tag) ); }
		
		/*
			Frees the memory allocated at the specified pointer
			pMemory: A pointer to the beginning of a block of memory allocated by the heap
			
			Optional Params (used for more verbose debugging data):
			line:	 line number of the call to free memory
			filename:filename of the line with the call to free memory
			
			throws MemoryException for failures
		*/
		void Free(void* pMemory, int line = 0, char* filename = 0);

		/*
			Creates a new Heap inside the memory allocated for the Heap
			size: Minimum requested size of the new Heap

			Optional Params (used for more verbose debugging data):
			line:	 line number of the call to create the heap
			filename:filename of the line with the call to create the heap
			
			throws MemoryException for failures
		*/
		Heap* SubHeap(size_t size, int line = 0, char* filename = 0);

		/*
			Frees the SubHeap pointed to by the provided pointer
			heap: Pointer to the Heap to free

			Optional Params (used for more verbose debugging data):
			line:	 line number of the call to free the heap
			filename:filename of the line with the call to free the heap
			
			throws MemoryException for failures
		*/
		void FreeSubHeap(Heap* heap, int line = 0, char* filename = 0);

		/* Merges block with sourrounding free blocks of memory. If pMemory==0, entire heap is coalesced */
		void Coalesce(void *pMemory = 0);	

		/* Checks all guard blocks in heap for changes. Returns true if Heap is valid, false otherwise */
		bool Validate();

		/* Returns size of all free memory if checkFree = true. Otherwise returns size of allocated memory */
		size_t TotalMemory(bool checkFree);

		/* Returns size of largest free block in the Heap */
		size_t LargestFreeBlock();

		/* Checks if the block is allocated by the Heap and checks if the block is corrupted*/
		bool IsInvalid(void *pMemory);

		/* Returns 0-1 value for fragmentation ratio. 0=no fragmentation 1=max fragmentation */
		float ExternalFragmentation();

		/* Returns 0-1 value for fragmentation ratio. 0=no fragmentation 1=max fragmentation */
		float InternalFragmentation();

		/* Returns the largest amount of memory ever stored in allocated blocks at once */
		int GetMaxMemoryUsage();

		/* prints recorded statistics for the heap */
		void PrintHeapStatistics();

		//=====================================================================
		//	Helper Debug Functions				  
		//=====================================================================
		
		/* Returns the number of unfreed blocks */
		int NumLeaks();

		/* Prints the allocation number, file, line, and tag (if provided) of any unallocated memory */
		void PrintLeaks();

		/* Returns oldest allocated block in debug. returns 0 in release */
		void* OldestBlock();

		/* Prints the number of bytes stored in blocks marked by the given tag */
		size_t GetUsageFromTag(int tag);

	private:
		void* Alloc(size_t size, int line = 0, char* filename = 0, int tag = 0);
		
		Heap_Impl* heap;
	};
}