#include "MemoryException.h"
#include "MemoryStructs.h"
#include <cassert>

namespace Combustion
{
const int _MAGIC = 0xBAADF00D;
#define GUARD_SIZE 4
#define TRUE_MIN_BLOCK_SIZE sizeof(MemoryHeader) + GUARD_SIZE + 4 
#define BLOCK_HEADER_AND_GUARD sizeof(MemoryHeader) + GUARD_SIZE

	struct MemoryHeader
	{
		#ifdef _DEBUG
				unsigned int allocNum;		//number of alloc (used to report unfreed memory)
				int lineNum;				//line that called Alloc
				int tag;					//block tag information
				char* file;					//file of line that called Alloc
		#endif

		size_t size;				//size of data in the block (not incl. guard block or header size)
		bool isAllocated;			//marks whether block is free or allocated
		MemoryHeader* prevBlock;	//pointer to prev block in heap
		MemoryHeader* nextBlock;	//pointer to next block in heap
		int guardData;				//contains begining guard block data

		MemoryHeader(size_t memsize, int guard, bool isalloced, int memoryTag = 0) 
			: size(memsize), isAllocated(isalloced), prevBlock(0), nextBlock(0), guardData(guard)
		{
			#ifdef _DEBUG
				tag = memoryTag;
			#endif
		}
	};

	struct Heap_Impl
	{
		size_t size;				//size of the entire heap
		int currUsage;				//current amount allocated
		int maxUsage;				//largest amount allocated
		int numAllocatedBlocks;		//number of currently allocated blocks
		int numTotalBlocks;			//number of all blocks (free and allocated)

		#ifdef _DEBUG
				unsigned int allocNum;		//allocation number
				HeapStats stats;			//structure that holds debugging statistics about heap usage
		#endif

		/*
			Constructs the Heap Header and creates a free block spanning all memory not taken up by the Heap Header
		*/
		Heap_Impl(size_t heapSize, int line = 0, char* filename = 0) 
			: size(heapSize - sizeof(Heap_Impl)), numAllocatedBlocks(0), numTotalBlocks(1), currUsage(0), maxUsage(0)
		{
			size_t minSize = sizeof(Heap_Impl) + sizeof(MemoryHeader);
			if(heapSize < minSize) throw MemoryException("Out of Memory", line, filename);

			//Create Header for a free block
			MemoryHeader* pFirstFree = (MemoryHeader*) (reinterpret_cast<char*>(this) + sizeof(Heap_Impl));
			pFirstFree = new(pFirstFree) MemoryHeader( heapSize - minSize, 0xBAADF00D, false );
			
			#ifdef _DEBUG
				pFirstFree->lineNum = line;
				pFirstFree->file = filename;
				stats = HeapStats();
			#endif
		}

		/*  Attempts to split a free block to make an allocated block of minSize and a free
			block for the remaining data.
			In the case where the resulting free block is too small to hold a header, the block
			is not split, and the allocated block is given the remainder.

			Allocated blocks are marked at the end with a guard block marker
		*/
		void Split(void* pHeader, size_t minSize)
		{
			MemoryHeader* memHeader = reinterpret_cast<MemoryHeader*>(pHeader);

			assert(memHeader->size >= minSize);
			assert(memHeader->isAllocated == false);

			//Check if there is enough leftover to justify split
			if(memHeader->size > TRUE_MIN_BLOCK_SIZE)
			{	//can split
				size_t originalSize = sizeof(MemoryHeader) + memHeader->size;

				//create first header
				memHeader->isAllocated = true;
				memHeader->size = minSize;
				//write magic marker at end
				int* ptr = reinterpret_cast<int*>( (reinterpret_cast<char*>(memHeader) + sizeof(MemoryHeader) + memHeader->size) );
				*ptr = _MAGIC;
				numAllocatedBlocks++;
				numTotalBlocks++;
				
				//create second header
				MemoryHeader* freeHeader = reinterpret_cast<MemoryHeader*>(
					(reinterpret_cast<char*>(memHeader) + sizeof(MemoryHeader) + memHeader->size + GUARD_SIZE) );
				
				int freeBlockSize = originalSize - (sizeof(MemoryHeader) * 2 + GUARD_SIZE + memHeader->size);
				freeHeader = new(freeHeader) MemoryHeader(freeBlockSize, _MAGIC, false);
				
				//repair pointers
				freeHeader->prevBlock = memHeader;
				freeHeader->nextBlock = memHeader->nextBlock;
				memHeader->nextBlock = freeHeader;
			}
			else
			{	//can't split, return entire block
				memHeader->isAllocated = true;
				int* ptr = (int*) (reinterpret_cast<char*>(memHeader) + sizeof(MemoryHeader) + memHeader->size);
				*ptr = _MAGIC;
				numAllocatedBlocks++;
			}
			currUsage += memHeader->size;
			if(currUsage > maxUsage) 
				maxUsage = currUsage;
		}

		/*
			Coalesces the block pointed to by pHeader with the block on its right
		*/
		void MergeWithNextBlock(void* pHeader)
		{
			MemoryHeader* currHeader = reinterpret_cast<MemoryHeader*>(pHeader);
			MemoryHeader* nextHeader = currHeader->nextBlock;

			size_t newSize = currHeader->size + sizeof(MemoryHeader) + nextHeader->size;
			nextHeader->guardData = _MAGIC;

			//clear the endGuard
			if(nextHeader->isAllocated)
			{
				int* endGuard = (int*) (reinterpret_cast<char*>(nextHeader) + sizeof(MemoryHeader) + nextHeader->size);
				*endGuard = _MAGIC;
				newSize += GUARD_SIZE;
			}

			currHeader->isAllocated = false;
			currHeader->size = newSize;
			currHeader->nextBlock = nextHeader->nextBlock;

			//fix the pointer of the next block
			if(nextHeader->nextBlock != 0)
				nextHeader->nextBlock->prevBlock = currHeader;

			//clear nextHeader's pointers for better debugging
			nextHeader->nextBlock = 0;
			nextHeader->prevBlock = 0;

			numTotalBlocks--;
		}
	};
}