#include <iostream>
#include "Heap.h"
#include "MemoryException.h"
#include <list>

using namespace std;

namespace Combustion
{
#define TEST(func)  if(func){ passed++; cout << "PASS\n"; }\
					else cout << "FAIL\n";\
					total++;
#define TRY try{
#define CATCH(output) }catch(MemoryException&){output}

const int NORMAL_ALLOC = 5120;
const int CRUEL_ALLOC = NORMAL_ALLOC * 100;

	class UnitTests
	{
	public:
		static void RunAllTests()
		{
			cout << "Unit Test Results\n------------------\n";

			srand( 0 );

			int passed = 0, total = 0;
		
			TEST( HeapInitTest() );	
			TEST( HeapBadInitTest() );
			TEST( AllocEmptyTest() );
			TEST( BadDeallocTest() );
			TEST( OutOfMemoryTest() );	
			TEST( SubHeapOutOfMemoryTest() );
			TEST( SubHeapFreeTest() );
			TEST( DoubleFreeTest() );
			TEST( MaxUsageTest() );
			TEST( CorruptedHeapTest() );
			TEST( CoalesceTest() );
			TEST( ReplaceTest() );
			TEST( DanglingPointerTest() );

			TEST( StressAllocDeallocTest() );
#ifdef _DEBUG
			TEST( TagTest() );
			TEST( MemoryLeakTest() );
#endif
			cout << endl << passed << " out of " << total << " tests passed" << endl;
			cout << "------------------\n\n";
		}

	private:
		
		/* Tests successful creation of a Heap, given valid memory */
		static bool HeapInitTest()
		{
			cout << "Heap Init Test Result:\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];

			TRY
				Heap heap = Heap(pMem, NORMAL_ALLOC);
				result = true;
			CATCH( result = false; )
		
			delete pMem;
			return result;
		}

		/* Tests successful MemoryException thrown when given too little memory during Heap creation */
		static bool HeapBadInitTest()
		{
			cout << "Heap Bad Init Test Result:\t\t";
			bool result;
			char* pMem = new char[0];

			TRY
				Heap heap = Heap(pMem, 0);
				result = false;
			CATCH( result = true; )
		
			delete pMem;
			return result;
		}
		
		/* Tests that allocating an empty block does not throw any exceptions or crash */
		static bool AllocEmptyTest()
		{
			cout << "Alloc Empty Test Result:\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);

			TRY
				void* empty = heap.Alloc<void>(0);
				heap.Free(empty);
				result = true;
			CATCH( result = false; )

			delete pMem;
			return result;
		}

		/* Tests freeing a non-valid pointer */
		static bool BadDeallocTest()
		{
			cout << "Bad Dealloc Test Result:\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);

			TRY
				void* empty = heap.Alloc<void>(100);
				
				void* invalidPtr = reinterpret_cast<void*>(reinterpret_cast<char*>(empty) + 50);

				heap.Free(invalidPtr);
				result = false;
			CATCH( result = true; )

			delete pMem;
			return result;
		}

		/* Tests successful exception when requesting more memory than the Heap can provide */
		static bool OutOfMemoryTest()
		{
			cout << "Out of Memory Test Result:\t\t";
			bool result;
			char* pMemory = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMemory, NORMAL_ALLOC);
			
			TRY
				heap.Alloc<int>(100);
				size_t size = heap.LargestFreeBlock();
				heap.Alloc<int>(size + 1);
				result = false;
			CATCH(result = true;)

			delete pMemory;
			return result;
		}

		/*  Tests correct counting of allocations/deallocations and reporting of
			un-freed memory */
		static bool MemoryLeakTest()
		{
			cout << "Memory Leak Test Result:\t\t";
			bool result;
			char* pMemory = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMemory, NORMAL_ALLOC);

			std::list<int*> blocks;

			for(int i = 0; i < 100; ++i)
				blocks.push_back( heap.Alloc<int>(10, __LINE__, __FILE__, 'test') );

			auto itr = blocks.begin();
			for(int i = 0; i < 99; ++i)
			{
				heap.Free((*itr));
				itr++;
			}

			if(heap.NumLeaks() == 1) 
				result = true;
			else
				result = false;

			//cout << endl;
			//heap.PrintLeaks();

			delete pMemory;
			return result;
		}
		
		/* Tests successful exception thrown when requesting more memory than a subheap has */
		static bool SubHeapOutOfMemoryTest()
		{
			cout << "SubHeap Out of Memory Test Result:\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);
			
			TRY
				heap.Alloc<int>(NORMAL_ALLOC - 200);
				heap.SubHeap(200);
				result = false;
			CATCH(result = true;)
		
			delete pMem;
			return result;
		}

		/* Tests successful free of subheap */
		static bool SubHeapFreeTest()
		{
			cout << "SubHeap Free Result:\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);
			
			int originalSize = heap.LargestFreeBlock();
			Heap* subheap = heap.SubHeap(NORMAL_ALLOC - 500);
			heap.FreeSubHeap(subheap);
			heap.Coalesce();

			if( heap.LargestFreeBlock() == originalSize )
				result = true;
			else
				result = false;

			delete pMem;
			return result;
		}
		
		/* Tests exception thrown for double free */
		static bool DoubleFreeTest()
		{
			cout << "Double Free Test Results:\t\t";
			bool result;
			char* pMemory = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMemory, NORMAL_ALLOC);

			void* pmem = heap.Alloc<int>(1024);
			heap.Free(pmem);
		
			TRY
				heap.Free(pmem);
				result = false;
			CATCH(result = true;)
		
			delete pMemory;
			return result;
		}
		
		/* Tests the validate function by corrupting the heap and calling validate */
		static bool CorruptedHeapTest()
		{
			cout << "Corrupted Heap Test:\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);

			int* ptr = heap.Alloc<int>(1024);

			int* hack = reinterpret_cast<int*>( reinterpret_cast<char*>(ptr) + 1024 );
			*hack = 0xABABABAB;

			if(heap.Validate())
				result = false;
			else
				result = true;
		
			delete pMem;
			return result;
		}
		
		/* Tests coalesce by allocating, then freeing data, and coalescing.
		   Checks starting and final free block size to test functionality */
		static bool CoalesceTest()
		{
			cout << "Coalesce Test\t\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);
		
			int freeStart = heap.TotalMemory(true);

			std::list<int*> blocks;

			for(int i = 0; i < 100; ++i)
				blocks.push_back( heap.Alloc<int>(1) );

			for(auto itr = blocks.begin(); itr != blocks.end(); ++itr)
				heap.Free( (*itr) );

			heap.Coalesce();

			if( freeStart == heap.TotalMemory(true) )
				result = true;
			else
				result = false;

			delete pMem;
			return result;
		}

		/* Tests that when a block of memory is freed a block with the same size can be
		   allocated in the same place */
		static bool ReplaceTest()
		{
			cout << "Replace Test\t\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);
		
			TRY
				int* first = heap.Alloc<int>(1500);
				int* second = heap.Alloc<int>(1500);
				int* third = heap.Alloc<int>(1500);
				heap.Free(second);
				int* fourth = heap.Alloc<int>(1500);

				result = true;
			CATCH( result = false; )

			delete pMem;
			return result;
		}

		/* Tests the IsInvalid method to ensure that pointers to free memory
		   can be detected */
		static bool DanglingPointerTest()
		{
			cout << "Dangling Pointer Test:\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);
		
			int* pMemory = heap.Alloc<int>(512);
			heap.Free(pMemory);
		
			if(heap.IsInvalid(pMemory) == true)
				result = true;
			else
				result = false;
		
			delete pMem;
			return result;
		}

		/* Tests that the total size for tagged blocks is correct*/
		static bool TagTest()
		{
			cout << "Tag Test:\t\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);

			for(int i = 0; i < 100; ++i)
			{
				int tag;
				(i % 2 == 0) ? tag = 'test' : tag = 'ntst';
				heap.Alloc<void>(10, 0, 0, tag);
			}

			if(heap.GetUsageFromTag('test') == 500)
				result = true;
			else
				result = false;

			delete pMem;
			return result;
		}

		/* Allocates data 5000 times, deallocating and coalescing when necessary 
		   Should not crash or throw exceptions! */
		static bool StressAllocDeallocTest()
		{
			cout << "Stress Alloc Dealloc Test:\t\t";
			bool result;
			char* pMem = new char[CRUEL_ALLOC];
			Heap heap = Heap(pMem, CRUEL_ALLOC);
			int size;

			std::list<int*> blocks;

			TRY
				//randomly allocate and deallocate blocks 5000 times
				for(int i = 0; i < 5000; ++i)
				{
					if( heap.LargestFreeBlock() >= 500 )
					{	//allocate
						(rand() % 2 == 0) ? size = 200 : size = 400;
						blocks.push_back( heap.Alloc<int>(size) );
					}
					else
					{	//deallocate
						if(blocks.size() == 0)
							heap.Coalesce();
						else
						{
							int index = rand() % blocks.size();
							auto itr = blocks.begin();

							while(index > 0) { ++itr; --index; }

							heap.Free((*itr));
							blocks.erase(itr);
						}
					}
				}

				//free remaining unfreed blocks
				for(auto itr = blocks.begin(); itr != blocks.end(); ++itr)
					heap.Free((*itr));
				blocks.clear();

				result = true;
			CATCH( result = false; )

			delete pMem;
			return result;
		}

		/* Tests the record for maximum memory allocated at once (worst case memory usage during run) */
		static bool MaxUsageTest()
		{
			cout << "Max Usage Test:\t\t\t\t";
			bool result;
			char* pMem = new char[NORMAL_ALLOC];
			Heap heap = Heap(pMem, NORMAL_ALLOC);
			
			void* first = heap.Alloc<void>(100);
			void* second = heap.Alloc<void>(100);
			void* third = heap.Alloc<void>(100);
			heap.Free(first);
			void* fourth = heap.Alloc<void>(100);
			
			if(heap.GetMaxMemoryUsage() == 300)
				result = true;
			else
				result = false;

			delete pMem;
			return result;
		}
	};
}